@@ -1,6 +1,7 @@
 # -*- coding: utf-8 -*-
 import json
 import traceback
+import warnings
 from collections import Counter
 from datetime import datetime
 
@@ -30,12 +31,14 @@ class Reporter(object):
                 "started": datetime.utcnow().isoformat(),
                 # Created element counts, by type slug
                 "elements": {},
-                # Created transcription counts, by type
-                "transcriptions": {},
+                # Created transcriptions count
+                "transcriptions": 0,
                 # Created classification counts, by class
                 "classifications": {},
                 # Created entities ({"id": "", "type": "", "name": ""}) from this element
                 "entities": [],
+                # Created metadata ({"id": "", "type": "", "name": ""}) from this element
+                "metadata": [],
                 "errors": [],
             },
         )
@@ -80,26 +83,35 @@ class Reporter(object):
         )
         element["classifications"] = dict(counter)
 
-    def add_transcription(self, element_id, type, type_count=1):
+    def add_transcription(self, element_id, type=None, type_count=None):
         """
         Report creating a transcription on an element.
-        Multiple transcriptions with the same type and parent can be declared with the type_count parameter.
+        Multiple transcriptions with the same parent can be declared with the type_count parameter.
         """
-        transcriptions = self._get_element(element_id)["transcriptions"]
-        transcriptions.setdefault(type, 0)
-        transcriptions[type] += type_count
+        if type_count is None:
+            if isinstance(type, int):
+                type_count, type = type, None
+            else:
+                type_count = 1
+
+        if type is not None:
+            warnings.warn(
+                "Transcription types have been deprecated and will be removed in the next release.",
+                FutureWarning,
+            )
+
+        self._get_element(element_id)["transcriptions"] += type_count
 
     def add_transcriptions(self, element_id, transcriptions):
         """
         Report one or more transcriptions at once.
         """
         assert isinstance(transcriptions, list), "A list is required for transcriptions"
-        element = self._get_element(element_id)
-        # Retrieve the previous existing transcription counts, if any
-        counter = Counter(**element["transcriptions"])
-        # Add the new ones
-        counter.update([transcription["type"] for transcription in transcriptions])
-        element["transcriptions"] = dict(counter)
+        warnings.warn(
+            "Reporter.add_transcriptions is deprecated due to transcription types being removed. Please use Reporter.add_transcription(element_id, count) instead.",
+            FutureWarning,
+        )
+        self.add_transcription(element_id, len(transcriptions))
 
     def add_entity(self, element_id, entity_id, type, name):
         """
@@ -114,6 +126,13 @@ class Reporter(object):
     def add_entity_role(self, *args, **kwargs):
         raise NotImplementedError
 
+    def add_metadata(self, element_id, metadata_id, type, name):
+        """
+        Report creating a metadata from an element.
+        """
+        metadata = self._get_element(element_id)["metadata"]
+        metadata.append({"id": metadata_id, "type": type, "name": name})
+
     def error(self, element_id, exception):
         error_data = {
             "class": exception.__class__.__name__,
