@@ -5,6 +5,7 @@ import logging
 import os
 import sys
 import uuid
+import warnings
 from enum import Enum
 from pathlib import Path
 
@@ -18,6 +19,8 @@ from arkindex_worker import logger
 from arkindex_worker.models import Element
 from arkindex_worker.reporting import Reporter
 
+MANUAL_SLUG = "manual"
+
 
 class BaseWorker(object):
     def __init__(self, description="Arkindex Base Worker"):
@@ -81,6 +84,11 @@ class BaseWorker(object):
         self.api_client = ArkindexClient(**options_from_env())
         logger.debug(f"Setup Arkindex API client on {self.api_client.document.url}")
 
+        # Load features available on backend, and check authentication
+        user = self.api_client.request("RetrieveUser")
+        logger.debug(f"Connected as {user['display_name']} - {user['email']}")
+        self.features = user["features"]
+
         if self.worker_version_id:
             # Retrieve initial configuration from API
             worker_version = self.api_client.request(
@@ -177,6 +185,22 @@ class EntityType(Enum):
     Date = "date"
 
 
+class MetaType(Enum):
+    Text = "text"
+    HTML = "html"
+    Date = "date"
+    Location = "location"
+    # Element's original structure reference (intended to be indexed)
+    Reference = "reference"
+
+
+class ActivityState(Enum):
+    Queued = "queued"
+    Started = "started"
+    Processed = "processed"
+    Error = "error"
+
+
 class ElementsWorker(BaseWorker):
     def __init__(self, description="Arkindex Elements Worker"):
         super().__init__(description)
@@ -242,13 +266,18 @@ class ElementsWorker(BaseWorker):
                     **self.api_client.request("RetrieveElement", id=element_id)
                 )
                 logger.info(f"Processing {element} ({i}/{count})")
+
+                # Report start of process, run process, then report end of process
+                self.update_activity(element, ActivityState.Started)
                 self.process_element(element)
+                self.update_activity(element, ActivityState.Processed)
             except ErrorResponse as e:
                 failed += 1
                 logger.warning(
                     f"An API error occurred while processing element {element_id}: {e.title} - {e.content}",
                     exc_info=e if self.args.verbose else None,
                 )
+                self.update_activity(element, ActivityState.Error)
                 self.report.error(element_id, e)
             except Exception as e:
                 failed += 1
@@ -256,6 +285,7 @@ class ElementsWorker(BaseWorker):
                     f"Failed running worker on element {element_id}: {e}",
                     exc_info=e if self.args.verbose else None,
                 )
+                self.update_activity(element, ActivityState.Error)
                 self.report.error(element_id, e)
 
         # Save report as local artifact
@@ -364,22 +394,36 @@ class ElementsWorker(BaseWorker):
 
         return sub_element["id"]
 
-    def create_transcription(self, element, text, type, score):
+    def create_transcription(self, element, text, type=None, score=None):
         """
-        Create a transcription on the given element through API
+        Create a transcription on the given element through the API.
+
+        Transcription types are deprecated; please call this method using
+        `create_transcription(element, text, score)` instead of
+        `create_transcription(element, text, type, score)`.
         """
         assert element and isinstance(
             element, Element
         ), "element shouldn't be null and should be of type Element"
-        assert type and isinstance(
-            type, TranscriptionType
-        ), "type shouldn't be null and should be of type TranscriptionType"
         assert text and isinstance(
             text, str
         ), "text shouldn't be null and should be of type str"
+
+        # When score is not set and type is not a transcription type, grab the score from `type`.
+        # Allows create_transcription(element, text, score) and (element, text, type, score)
+        # for forwards compatibility after transcription types get deleted.
+        if score is None and not isinstance(type, TranscriptionType):
+            score, type = type, None
+        elif isinstance(type, TranscriptionType):
+            warnings.warn(
+                "Transcription types are deprecated and will be removed in the next release.",
+                FutureWarning,
+            )
+
         assert (
             isinstance(score, float) and 0 <= score <= 1
         ), "score shouldn't be null and should be a float in [0..1] range"
+
         if self.is_read_only:
             logger.warning(
                 "Cannot create transcription as this worker is in read-only mode"
@@ -391,12 +435,11 @@ class ElementsWorker(BaseWorker):
             id=element.id,
             body={
                 "text": text,
-                "type": type.value,
                 "worker_version": self.worker_version_id,
                 "score": score,
             },
         )
-        self.report.add_transcription(element.id, type.value)
+        self.report.add_transcription(element.id)
 
     def create_classification(
         self, element, ml_class, confidence, high_confidence=False
@@ -493,10 +536,14 @@ class ElementsWorker(BaseWorker):
         return entity["id"]
 
     def create_element_transcriptions(
-        self, element, sub_element_type, transcription_type, transcriptions
+        self, element, sub_element_type, transcription_type=None, transcriptions=None
     ):
         """
         Create multiple sub elements with their transcriptions on the given element through API
+
+        Transcription types are deprecated; please call this method using
+        `create_element_transcriptions(element, sub_element_type, transcriptions)` instead of
+        `create_element_transcriptions(element, sub_element_type, transcription_type, transcriptions)`.
         """
         assert element and isinstance(
             element, Element
@@ -504,9 +551,20 @@ class ElementsWorker(BaseWorker):
         assert sub_element_type and isinstance(
             sub_element_type, str
         ), "sub_element_type shouldn't be null and should be of type str"
-        assert transcription_type and isinstance(
+
+        # When transcriptions are not set and transcription_type is not a valid transcription type,
+        # take transcriptions from `transcription_type`.
+        # Allows for forwards compatibility after transcription types get deleted.
+        if transcriptions is None and not isinstance(
             transcription_type, TranscriptionType
-        ), "transcription_type shouldn't be null and should be of type TranscriptionType"
+        ):
+            transcriptions, transcription_type = transcription_type, None
+        elif isinstance(transcription_type, TranscriptionType):
+            warnings.warn(
+                "Transcription types are deprecated and will be removed in the next release.",
+                FutureWarning,
+            )
+
         assert transcriptions and isinstance(
             transcriptions, list
         ), "transcriptions shouldn't be null and should be of type list"
@@ -546,7 +604,6 @@ class ElementsWorker(BaseWorker):
             id=element.id,
             body={
                 "element_type": sub_element_type,
-                "transcription_type": transcription_type.value,
                 "worker_version": self.worker_version_id,
                 "transcriptions": transcriptions,
                 "return_elements": True,
@@ -558,10 +615,47 @@ class ElementsWorker(BaseWorker):
                     f"A sub_element of {element.id} with type {sub_element_type} was created during transcriptions bulk creation"
                 )
                 self.report.add_element(element.id, sub_element_type)
-            self.report.add_transcription(annotation["id"], transcription_type.value)
+            self.report.add_transcription(annotation["id"])
 
         return annotations
 
+    def create_metadata(self, element, type, name, value, entity=None):
+        """
+        Create a metadata on the given element through API
+        """
+        assert element and isinstance(
+            element, Element
+        ), "element shouldn't be null and should be of type Element"
+        assert type and isinstance(
+            type, MetaType
+        ), "type shouldn't be null and should be of type MetaType"
+        assert name and isinstance(
+            name, str
+        ), "name shouldn't be null and should be of type str"
+        assert value and isinstance(
+            value, str
+        ), "value shouldn't be null and should be of type str"
+        if entity:
+            assert isinstance(entity, str), "entity should be of type str"
+        if self.is_read_only:
+            logger.warning("Cannot create metadata as this worker is in read-only mode")
+            return
+
+        metadata = self.api_client.request(
+            "CreateMetaData",
+            id=element.id,
+            body={
+                "type": type.value,
+                "name": name,
+                "value": value,
+                "entity": entity,
+                "worker_version": self.worker_version_id,
+            },
+        )
+        self.report.add_metadata(element.id, metadata["id"], type.value, name)
+
+        return metadata["id"]
+
     def get_worker_version(self, worker_version_id: str) -> dict:
         """
         Get worker version from cache if possible, otherwise make API request
@@ -604,5 +698,143 @@ class ElementsWorker(BaseWorker):
         # transcriptions have worker_version_id but elements have worker_version
         elif "worker_version_id" in ml_result and ml_result["worker_version_id"]:
             return self.get_worker_version_slug(ml_result["worker_version_id"])
+        elif "worker_version" in ml_result and ml_result["worker_version"] is None:
+            return MANUAL_SLUG
+        elif (
+            "worker_version_id" in ml_result and ml_result["worker_version_id"] is None
+        ):
+            return MANUAL_SLUG
         else:
             raise ValueError(f"Unable to get slug from: {ml_result}")
+
+    def list_transcriptions(
+        self, element, element_type=None, recursive=None, worker_version=None
+    ):
+        """
+        List transcriptions on an element
+        """
+        assert element and isinstance(
+            element, Element
+        ), "element shouldn't be null and should be of type Element"
+        query_params = {}
+        if element_type:
+            assert isinstance(element_type, str), "element_type should be of type str"
+            query_params["element_type"] = element_type
+        if recursive is not None:
+            assert isinstance(recursive, bool), "recursive should be of type bool"
+            query_params["recursive"] = recursive
+        if worker_version:
+            assert isinstance(
+                worker_version, str
+            ), "worker_version should be of type str"
+            query_params["worker_version"] = worker_version
+
+        transcriptions = self.api_client.paginate(
+            "ListTranscriptions", id=element.id, **query_params
+        )
+
+        return transcriptions
+
+    def list_element_children(
+        self,
+        element,
+        best_class=None,
+        folder=None,
+        name=None,
+        recursive=None,
+        type=None,
+        with_best_classes=None,
+        with_corpus=None,
+        with_has_children=None,
+        with_zone=None,
+        worker_version=None,
+    ):
+        """
+        List children of an element
+        """
+        assert element and isinstance(
+            element, Element
+        ), "element shouldn't be null and should be of type Element"
+        query_params = {}
+        if best_class is not None:
+            assert isinstance(best_class, str) or isinstance(
+                best_class, bool
+            ), "best_class should be of type str or bool"
+            query_params["best_class"] = best_class
+        if folder is not None:
+            assert isinstance(folder, bool), "folder should be of type bool"
+            query_params["folder"] = folder
+        if name:
+            assert isinstance(name, str), "name should be of type str"
+            query_params["name"] = name
+        if recursive is not None:
+            assert isinstance(recursive, bool), "recursive should be of type bool"
+            query_params["recursive"] = recursive
+        if type:
+            assert isinstance(type, str), "type should be of type str"
+            query_params["type"] = type
+        if with_best_classes is not None:
+            assert isinstance(
+                with_best_classes, bool
+            ), "with_best_classes should be of type bool"
+            query_params["with_best_classes"] = with_best_classes
+        if with_corpus is not None:
+            assert isinstance(with_corpus, bool), "with_corpus should be of type bool"
+            query_params["with_corpus"] = with_corpus
+        if with_has_children is not None:
+            assert isinstance(
+                with_has_children, bool
+            ), "with_has_children should be of type bool"
+            query_params["with_has_children"] = with_has_children
+        if with_zone is not None:
+            assert isinstance(with_zone, bool), "with_zone should be of type bool"
+            query_params["with_zone"] = with_zone
+        if worker_version:
+            assert isinstance(
+                worker_version, str
+            ), "worker_version should be of type str"
+            query_params["worker_version"] = worker_version
+
+        children = self.api_client.paginate(
+            "ListElementChildren", id=element.id, **query_params
+        )
+
+        return children
+
+    def update_activity(self, element, state):
+        """
+        Update worker activity for this element
+        This method should not raise a runtime exception, but simply warn users
+        """
+        assert element and isinstance(
+            element, Element
+        ), "element shouldn't be null and should be of type Element"
+        assert isinstance(state, ActivityState), "state should be an ActivityState"
+
+        if not self.features.get("workers_activity"):
+            logger.debug("Skipping Worker activity update as it's disabled on backend")
+            return
+
+        if self.is_read_only:
+            logger.warning("Cannot update activity as this worker is in read-only mode")
+            return
+
+        try:
+            out = self.api_client.request(
+                "UpdateWorkerActivity",
+                id=self.worker_version_id,
+                body={
+                    "element_id": element.id,
+                    "state": state.value,
+                },
+            )
+            logger.debug(f"Updated activity of element {element.id} to {state}")
+            return out
+        except ErrorResponse as e:
+            logger.warning(
+                f"Failed to update activity of element {element.id} to {state.value} due to an API error: {e.content}"
+            )
+        except Exception as e:
+            logger.warning(
+                f"Failed to update activity of element {element.id} to {state.value}: {e}"
+            )
